#!/usr/bin/perl -w

use strict;

use DBI;
use Digest::MD5 qw(md5_hex);
use File::Path qw(remove_tree);
use HTML::Parser;
require LWP::UserAgent;

die "redsnapper olddir newdir" unless @ARGV >= 2;

my $snapnum;

my $tmpdir = "/tmp/redsnap" . int(rand(9999));
my %inodes;

my $dbh;
my $diffstmt;
my $dirstmt;
my $filestmt;
my $linkstmt;

my $pkgmir = "";

sub fixname {
	my $s = shift;

	return substr($s, 3);
}

sub readall {
	my $file = shift;
	my @info = stat $file or die "where'd $file go?";
	my $size = $info[7];
	my $fd;

	open $fd, $file;

	my $data;
	read $fd, $data, $size;
	close $fd;
	return $data;
}

sub writeall {
	my $file = shift;
	my $data = shift;

	my $fd;
	open $fd, ">$file" or die "can't open $file";
	syswrite $fd, $data;
	close $fd;
}

sub httpget {
	my $url = shift;
	my $ua = LWP::UserAgent->new;
	$ua->default_header('Accept-Encoding' =>
		scalar HTTP::Message::decodable());
	my $resp = $ua->get($url);
	die "request to $url failed" unless $resp->is_success;
	return $resp->decoded_content;
}

sub diff {
	my $filea = shift;
	my $fileb = shift;
	my $difffile = "$tmpdir/difffile";

	unlink $difffile;
	if ($filea =~ /.*\.gz$/ || $filea =~ /.*\.tgz$/) {
		system "gzcat $filea > $tmpdir/ungza";
		system "gzcat $fileb > $tmpdir/ungzb";
		system "bsdiff $tmpdir/ungza $tmpdir/ungzb $difffile";
		unlink "$tmpdir/ungza", "$tmpdir/ungzb";
	} else {
		system "bsdiff $filea $fileb $difffile";
	}
	my $d = readall $difffile;
	unlink $difffile;
	return $d;
}

sub savefile {
	my $file = shift;

	my @info = stat $file;
	my $uid = $info[4];
	my $gid = $info[5];
	my $mode = $info[2];
	my $data = readall $file;
	$filestmt->execute(fixname($file), $data, $uid, $gid, $mode);
}

sub savedir {
	my $dir = shift;
	my @info = stat $dir;

	$dirstmt->execute(fixname($dir), $info[4], $info[5], $info[2]);
}

my $savedfingers = 0;
sub savefinger {
	my $file = shift;
	my $md5 = shift;

	if ($file eq "/bsd" || $file eq "/bsd.mp") {
		return;
	}

	if ($savedfingers++ > 4) {
		return;
	}

	my $fd;
	open $fd, ">>", "snap-fingers.txt" or die "no snap finger";
	print $fd "$file $md5 $snapnum\n";
	close $fd;
}

sub process {
	my $patha = shift;
	my $pathb = shift;

	my $dir;
	if (! -d $patha) {
		savedir $pathb;
	}
	opendir $dir, $pathb;
	for my $ent (readdir $dir) {
		next if ($ent eq "." || $ent eq "..");

		my $filea = "$patha/$ent";
		my $fileb = "$pathb/$ent";
		my @info = lstat $fileb;

		if (-d _) {
			process($filea, $fileb);
		} elsif (-l _) {
			my $targ = readlink $fileb;
			$linkstmt->execute(fixname($fileb), $targ, 1);
		} elsif (-f _) {
			my $inode = $info[1];
			if ($inodes{$inode}) {
				$linkstmt->execute(fixname($fileb),
					$inodes{$inode}, 0);
				next;
			} else {
				$inodes{$inode} = fixname($fileb);
			}
			if (-f $filea) {
 				if (system "cmp $filea $fileb > /dev/null") {
					my $d = diff $filea, $fileb;
					my $m = md5_hex(readall $filea);
					savefinger fixname($filea), $m;
					$diffstmt->execute(fixname($fileb),
						$d, $m);
				}
			} else {
				savefile $fileb;
			}
		}
	}
	closedir $dir;
}

my %availpkgs;
my %oldpkgs;

sub pkgname {
	my $tarname = shift;
	my $pname = "";

	if ($tarname =~ /(([^-]|-[[:alpha:]])*)-(\d\.\d)?.*?(gtk2|no_x11)?\.tgz/) {
		$pname = $1;
		my $flav = $4;
		if ($pname eq "python") {
			$pname .= "-$3";
		}
		$pname .= "-$flav" if $flav;
	}
	return $pname;
}

sub starttag {
	my $attrs = shift;

	my $pfile = $$attrs{'href'};
	my $pname = pkgname $pfile;
	$availpkgs{$pname} = $pfile;

}

sub getavailpkgs {
	my $idx = readall "index.html"; #httpget "$pkgmir/";

	my $parser = HTML::Parser->new(api_version => 3);
	$parser->handler(start => \&starttag, 'attr' );
	$parser->report_tags([ 'a' ]);
	$parser->parse($idx);
	$parser->eof;
}

sub getoldpkgs {
	my $patha = shift;
	my $hasnew = 0;

	my $dir;
	opendir $dir, $patha;
	for my $ent (readdir $dir) {
		next if ($ent eq "." || $ent eq "..");

		my $pname = pkgname $ent;
		$oldpkgs{$pname} = $ent;

		my $newpkg = $availpkgs{$pname};
		if ($newpkg && $newpkg ne $ent) {
			$hasnew = 1;
		}
	}
	closedir $dir;

	return $hasnew;
}

sub downloadnewpkgs {
	my $pathb = shift;

	my $curdir;
	opendir $curdir, ".";

	chdir $pathb;
	my $dir;
	opendir $dir, ".";
	my @tounlink = [];
	for my $ent (readdir $dir) {
		next if ($ent eq "." || $ent eq "..");

		my $pname = pkgname $ent;

		my $newpkg = $availpkgs{$pname};
		system "wget -N $pkgmir/$newpkg" if $newpkg;
		if ($newpkg && $newpkg ne $ent) {
			push @tounlink, $ent;
		}
	}
	for my $ent (@tounlink) {
		unlink $ent;
	}
	closedir $dir;
	chdir $curdir;
	closedir $curdir;
}

sub diffpkgs {
	my $patha = shift;
	my $pathb = shift;
	my $hasnew;

	my $pkgnum = int(readall "pkgnum.txt");
	$pkgnum++;

	getavailpkgs;

	$hasnew = getoldpkgs $patha;
	return unless $hasnew;

	#downloadnewpkgs $pathb;

	my $fd;
	open $fd, ">>", "pkgs.txt" or die "no pkgs.txt";

	my $dir;
	opendir $dir, $pathb;
	for my $ent (readdir $dir) {
		next if ($ent eq "." || $ent eq "..");

		my $pname = pkgname $ent;
		my $fileb = "$pathb/$ent";
		my $oldpkg = $oldpkgs{$pname};
		next unless $oldpkg;
		my $filea = "$patha/$oldpkg";

		if (system "cmp $filea $fileb > /dev/null") {
			my $d = diff $filea, $fileb;
			writeall "$pname-$pkgnum.bdiff", $d;
			my $md5 = md5_hex(readall $fileb);
			print $fd "$ent $md5 $pkgnum\n";
		}
	}
	closedir $dir;
	close $fd;

	writeall "pkgnum.txt", $pkgnum;

	remove_tree($patha);
	system "cp -pr $pathb $patha";

}


sub main {
	mkdir $tmpdir, 0700 or die "can't make tmpdir $tmpdir";

	$snapnum = int(readall "latest.txt");

	unlink "diff.db";

	$dbh = DBI->connect("dbi:SQLite:diff.db") or die "no db";
	$dbh->do("create table diffs (file TEXT, diff BLOB, md5 TEXT)");
	$dbh->do("create table dirs (dir TEXT, uid INTEGER, gid INTEGER, " .
		"mode INTEGER)");
	$dbh->do("create table files (file TEXT, data BLOB, uid INTEGER, " .
		"gid INTEGER, mode INTEGER)");
	$dbh->do("create table links (src TEXT, dst TEXT, sym INTEGER)");
	$dbh->begin_work;
	$diffstmt = $dbh->prepare("insert into diffs (file, diff, md5) " .
		"values (?, cast(? as blob), ?)");
	$dirstmt = $dbh->prepare("insert into dirs (dir, uid, gid, mode) " .
		"values (?, ?, ?, ?)");
	$filestmt = $dbh->prepare("insert into files (file, data, uid, gid, " .
		"mode) values (?, cast(? as blob), ?, ?, ?)");
	$linkstmt = $dbh->prepare("insert into links (src, dst, sym) values " .
		"(?, ?, ?)");

	process $ARGV[0], $ARGV[1];

	$dbh->commit;
	$dbh->disconnect;

	$snapnum++;
	writeall "latest.txt", $snapnum;
	rename "diff.db", "diff$snapnum.db";

	if ($ARGV[2] && $ARGV[3]) {
		diffpkgs $ARGV[2], $ARGV[3];
	}

	remove_tree($tmpdir);


}

main;
